/**

Spending vs. Media Coverage
May, 2010
by: Pitch Interactive, Inc. (www.pitchinteractive.com)

Spending vs. Media Coverage is a visual exploration that depicts federal funding for agencies and
associates this relationship to media coverage. 

Dependencies:

- You will also need an NYT Article API Key (http://developer.nytimes.com/)
  With the API key replace the apiKey String below to run this file. 

Thanks to Jer Thorp (http://blog.blprnt.com/) for simplifying the NYT API calls for us and for his JSON Lib.

**/
import processing.pdf.*;

import processing.opengl.*;
import org.json.*;

String[] agencies = { //used to identify file names
  "agriculture", "commerce", "defense", "education", "energy", 
  "epa","genServicesAdmin", "health", "homelandSec", "housing", "interior", 
  "intlDevelop", "justice","labor","nasa","other","smallBusinessAdmin","socialSecAdmin",
  "state", "transportation", "treasury", "vetsAffairs"};
ArrayList data;
ArrayList allProducts;
ArrayList allContractors;
ArrayList lineData;
ArrayList textData;
String baseURL = "http://api.nytimes.com/svc/search/v1/article";                  //Base URL for the NYT Article Search API
String apiKey = "YOUR NYT ARTICLE API KEY GOES HERE";                          //Enter your API Key here (http://developer.nytimes.com/)
String facetString = "org_facet";                                                 //Type(s) of facet to include in the search
String xmlPath = "../data/";

int maxVal = 0;                                                                    //Keeps track of the maximum returned value over the all terms
int localMax = 0;                                                                  //Keeps track of the maximum returned value over the each term
int totalMonths = 0;
int lastTotal = 0;  
int s = 2009;                                                                      //Start Year
int e = 2010;                                                                      //End Year

/*next 5 values control size of arcs and text*/
float minArcSize = 30;
float maxArcSize = 250;
float initArcSize = 600;

int minFontSize = 6;
int maxFontSize = 75;

/*
Note: we're fully aware these are general words and that we could have expanded this further. 
while we wanted to add words like 'war' to defense, we could also add 'disease', 'cancer', 'hostpital' to health, 
and how to equate these to social security or dept. of interior? with this in mind, we decided to keep our search
words very general and use this as a guage to get an idea.
*/
String[] words = {"Defense", "Veteran","NASA",  "Education","Health","Interior","Energy",  "\"Homeland%20Security\"","Agriculture","Commerce","Environment","Housing",
  "\"International%20Development\"","Justice","Labor","\"Small%20Business\"","\"Social%20Security\"","\"State%20Department\"","Transportation","Treasury",
  "\"General%20Service\""};             //Search Terms
int[] wordTotals;
color backColor = #F2F2F2;

color c1 = color(0, 102, 153);
color c2 = color(204, 102, 0);
int total = 0;                                                                     //store the cumulative total of contract spending
boolean refresh = false;                                                           //Determines wether each term has it's own max value
boolean readyToDraw = false;                                                      //used to determine when all the data has been loaded properly
PFont font;
ArrayList keywordData;

/**
setup()
performs general initialization. Loads the data and does some brief analysis on it.

note the loadData() function loads data from XML, where as the NYT data is loaded through the buildMonthArray() function
*/
void setup() {
  size(2400,1000);
  textMode(SHAPE);
  hint(ENABLE_NATIVE_FONTS);  
  loadData();
  //Load the font to be used in text display
  lineData = new ArrayList();
  textData = new ArrayList();
  font = loadFont("TitilliumMaps26L-250wt-36.vlw"); 
  textFont(font); 
  
  //Set the size of the stage & set the background
  frameRate(1);
  background(backColor);
  smooth();
  keywordData = new ArrayList();
  //Draw key
  textSize(36);
  
  wordTotals = new int[words.length];
  TimesDataChunk[][] data = new TimesDataChunk[words.length][12];
  //For each keyword, request a TimesDataChunk and draw it.
  
  for (int i = 0; i < words.length; i++) {
    localMax = 0;
    totalMonths = 0;
    TimesDataChunk[] months = buildMonthArray(words[i], s, e);
    data[i] = months;
    int keywordTotal = 0;
    for(int j = 0; j < months.length; j++) {
      keywordTotal += months[j].total;
    }
    wordTotals[i] += keywordTotal;
    total += keywordTotal;
    KeywordData kd = new KeywordData();
    kd.keyword = words[i];
    kd.amount = keywordTotal;
    keywordData.add(kd);
    if (refresh) maxVal = 0;
  };
  Comparator c = new KeywordDataComparator();
  Collections.sort(keywordData,c);

  readyToDraw = true;
  //Save out an image when the whole process is finished
  save(words[0] + "_" + s + "_" + e + ".png");
}
void draw() {
  if(!readyToDraw) { //if the data has not been loaded yet, just wait for it
    return;
  }
  colorMode(HSB);
  background(255);
  strokeCap(SQUARE);
  textFont(font,10);
  
  beginRecord(PDF,"output.pdf"); //comment this line will speed up output if you don't want pdf generation
  /*next section draws the arcs and plots the text for the spending data */
  
  //quick hack to scale things up, make them a bit bigger.
  pushMatrix();
  translate(width/2,height/2);
  scale(1.5,1.5);
  translate(-width/2,-height/2);

  //define the center of the ring
  float ringCenterX = width/3.0;
  float ringCenterY = height/2.0;
  float arcSize = initArcSize;
  
  float arcDegs = PI/2;
  float totalFrac = 0;
  float totalSpending = 0;
  //first loop through the data once to get the total contract spending
  for(int i = 0; i < data.size(); i++) {
    SpendingData sp = (SpendingData)data.get(i);
    totalSpending += sp.totalObligation;
    
  }
  totalFrac = 0;
  
  /*
  Loop through a second time to compare each dept to the total
  The depts color is generated by interpolating between the two colors c1, and c2
  */
  for(int i = 0; i < data.size(); i++) {
    SpendingData sp = (SpendingData)data.get(i);
    float spent = sp.totalObligation;
    color c = lerpColor(c1,c2,((float)i)/((float)data.size()));
    totalFrac += spent / totalSpending;
    sp.col = c;
    noFill();
    stroke(c);
    //determine the width of this arc
    float arcWidth = map(sp.totalObligation, 0 ,totalSpending,0, maxArcSize);
    arcSize -= arcWidth; //subtract it from arcsize for next iteration
    
    //LineData is used later to draw the bezier curves. It is just used to keep track of the positions of each of the arcs
    LineData ld = new LineData();
    ld.dept = sp.dept; //department
    ld.col = c; //color
    ld.leftXPos = ringCenterX; //x position is constant for the ring
    ld.leftYPosStart = ringCenterY - arcSize/2.0 - arcWidth/2.0; //we would like the start Y Pos and the 
    ld.leftYPosEnd = ringCenterY - arcSize/2.0 + arcWidth/2.0; //end Y pos
    //technically could just save one and then just use the previous or next LineData's yPos as the other value
    //but this is much easier.
    lineData.add(ld); //add it to the lineData array
    
    /* instead of drawing the arcs with solid fills, here we draw them with a series of lines
       that together makes up the full arc */
    for(int j = 0; j < arcWidth; j++) {
      float startPos = PI+PI/2;
      float endPos = random(-PI/48/4,PI/48/4);
      float lineRadius = map(j,0,arcWidth,arcSize - arcWidth, arcSize+arcWidth);
      
      arc(ringCenterX, ringCenterY, lineRadius,lineRadius,endPos,startPos);
    }

    noStroke();
    fill(c);
    //finally draw the label for this department
    float fontSize = 10;
    fontSize = map(spent,0,totalSpending,minFontSize,maxFontSize);
    textFont(font,fontSize);
    
    float lblXCorner = ringCenterX;
    float lblYCorner = ringCenterY - arcSize/2.0 + fontSize/2.0;
    fill(0);
    
    //however we do actually want the text to appear on top of the bezier curves so just store it in an array of TextData objects to be drawn later
    TextData td = new TextData();
    td.theText = sp.dept;
    td.fontSize = fontSize;
    td.x = lblXCorner;
    td.y = lblYCorner;
    textData.add(td);
    //update a few things for the next iteration
    arcSize -= arcWidth;
    arcDegs += PI/data.size();
  }
  popMatrix();
  
  
  /* Below draws the arc for NYT media coverage. It works very similarly to the Spending Arc code */
  pushMatrix();
  translate(width/2,height/2);
  scale(1.5,1.5);
  translate(-width/2,-height/2);
  colorMode(HSB);
  strokeCap(SQUARE);
  textFont(font,10);
  totalSpending = total;
  ringCenterX = 2*width/3.0 ;
  ringCenterY = height/2.0;
  arcSize = initArcSize;
  
  arcDegs = PI/2;
  totalFrac = 0;
  for(int i = 0; i < keywordData.size(); i++) {
    KeywordData sp = (KeywordData)keywordData.get(i);
    float spent = sp.amount;
    color c = getCorrespondingSpendingColor(sp.keyword); //grab the color from the LineData based on this keyword
    totalFrac += spent / totalSpending;
    noFill();
    stroke(c);
    float arcWidth = map(spent, 0 ,totalSpending,0, maxArcSize);
    arcSize -= arcWidth;
    LineData ld = getCorrespondingLineData(sp.keyword); //grabs the corresponding LineData object that was created earlier. 
    ld.rightXPos = ringCenterX;
    ld.rightYPosStart = ringCenterY - arcSize/2.0 - arcWidth/2.0;
    ld.rightYPosEnd = ringCenterY - arcSize/2.0 + arcWidth/2.0;
    
    for(int j = 0; j < arcWidth; j++) {
      float startPos = PI+PI/2.0;
      float endPos = random(TWO_PI+PI-PI/48/4,TWO_PI+PI);
      float lineRadius = map(j,0,arcWidth,arcSize - arcWidth, arcSize+arcWidth);
      
      arc(ringCenterX, ringCenterY, lineRadius,lineRadius,startPos,endPos);
      
    }

    noStroke();
    fill(c);
    
    
    //again text is last, so we just store it and draw it later
    float fontSize = 10;
    fontSize = map(sp.amount,0,totalSpending,minFontSize,maxFontSize);
    textFont(font,fontSize);
    
    float lblXCorner = ringCenterX;
    float lblYCorner = ringCenterY - arcSize/2.0 + fontSize/2.0;
    fill(0);
    String lblText = sp.keyword.replace("%20"," ").replace("\"","");
    lblXCorner -= textWidth(lblText);
    println(lblText + ": "+spent);
    TextData td = new TextData();
    td.theText = lblText;
    td.fontSize = fontSize;
    td.x = lblXCorner;
    td.y = lblYCorner;
    textData.add(td);
    arcSize -= arcWidth;
    arcDegs += PI/keywordData.size();
  }
  
  popMatrix();  

  
  /**
  draw lines between regions
  */
  pushMatrix();
  translate(width/2,height/2);
  scale(1.5,1.5);
  translate(-width/2,-height/2);
  //for each of our lineData objects, there is one for each department
  for(int i = 0; i < lineData.size(); i++) {
    noFill();
    LineData ld = (LineData) lineData.get(i);
    
    stroke(ld.col,40); //use the lineData color but make it somewhat transparent
    float leftDiff = abs(ld.leftYPosStart - ld.leftYPosEnd);
    float rightDiff = abs(ld.rightYPosStart - ld.rightYPosEnd);
    float maxDiff = max(leftDiff,rightDiff);
    //whichever side of the line data has a larger value use that amount.
    //this keeps the bezier curves from looking fragmented.
    for(int j = 0; j < maxDiff; j++) {
      //find the anchor points for the current bezier segment
      float lX = ld.leftXPos;
      float lY = map(j, 0, maxDiff, ld.leftYPosStart, ld.leftYPosEnd);
      float rX = ld.rightXPos;
      float rY = map(j, 0 , maxDiff, ld.rightYPosStart, ld.rightYPosEnd);
      if(rX == 0 || rY == 0) { //if a dept has no results, don't use it
        continue;
      }
      float lowValue = 1.0/3.0; //pretty simple interpolation for bezier control points
      float highValue = 2.0/3.0;
      bezier(lX,lY, lX*highValue+rX*lowValue,lY,lX*lowValue+rX*highValue,rY,rX,rY);
    } 
  }
  popMatrix();
  
  
  
  /**
  finally draw the labels on top
  */
  pushMatrix();
  translate(width/2,height/2);
  scale(1.5,1.5);
  translate(-width/2,-height/2);
  for(int i = 0; i < textData.size(); i++) {
    TextData td = (TextData) textData.get(i);
    textFont(font, td.fontSize);
    text(td.theText, td.x, td.y);
  }
  
  popMatrix();
  endRecord();
  saveFrame("iter3.png");
  readyToDraw = false;
};

/*

buildMonthArray
- Makes a series of calls to getChunk to fill an array with TimesDataChunks
- Formats dated in required way

*/

TimesDataChunk[] buildMonthArray(String word, int startYear, int endYear) {
  //Find out how many months are going to be stored
  int months = (endYear - startYear) * 12;
  TimesDataChunk[] mArray = new TimesDataChunk[months];
  for (int j = 0; j < months; j++) {
   
    int i = j % 12;
    int y = startYear +  floor(j / 12);                                //GET YEAR
    String m = (i + 1 < 10) ? ("0" + str(int(i) + 1)):str(i + 1);      //GET MONTH
    
    totalMonths ++;
    
    TimesDataChunk dc2 = getChunk(word, str(y), m);                    //REQUEST DATA CHUNK
    mArray[j] = dc2;                                                   //ADD DATA CHUCK TO MONTH ARRAY
    
    //println ("PROCESSED " + j + " OF " + months);
  };
  return(mArray);
};
/*
returns the color for a given media keyword
*/
color getCorrespondingSpendingColor(String keyword) {
  String word = keyword.replace("%20"," ").replace("\"","");
  if(word.equals("State Department")) {
    word = "State";
  }
  for(int i = 0 ; i < data.size(); i++) {
   SpendingData sd = (SpendingData)data.get(i);
   if(sd.dept.contains(word)) {
     return sd.col;
   }
  }
  println("no color for "+keyword);
  return color(0);
}

/*
returns the LineData object for a given media keyword
*/
LineData getCorrespondingLineData(String keyword) {
  String word = keyword.replace("%20"," ").replace("\"","");
  if(word.equals("State Department")) {
    word = "State";
  }
  for(int i = 0; i < lineData.size(); i++) {
    LineData ld = (LineData)lineData.get(i);
    if(ld.dept.contains(word)) {
      return ld;
    }
  }  
  println("no line data for "+keyword);
  return null;
}
/*

getChunk
- Makes a request to the NYT Article Search API and returns a TimessDataChunk

*/

TimesDataChunk getChunk(String word, String y, String m) {
  String url = baseURL + "?query=" + word + "%20publication_year:[" + y + "]%20publication_month:[" + m + "]&fields=+&facets=" + facetString + "&api-key=" + apiKey;
  //println(url);
  
  TimesDataChunk dc = new TimesDataChunk();
  try { 
    JSONObject nytData = new JSONObject(join(loadStrings(url), ""));  
    JSONObject facets = nytData.getJSONObject("facets");
    dc.facets = facets.getJSONArray("org_facet");
    
    dc.facetStrings = new String[dc.facets.length()];
    for (int i = 0; i < dc.facets.length(); i++) {
      JSONObject o = (JSONObject) dc.facets.get(i);
      dc.facetStrings[i] = o.getString("term");
    };
  
    dc.total = int(nytData.getInt("total")); 
//    println("TOTAL FOR: "+word+" : "+dc.total);
    if (dc.total > localMax) {
      localMax = min(dc.total, 3000);
      if (localMax > maxVal) maxVal = localMax;
      
    };
    
    if (dc.total > float(lastTotal) * 1.2) {
      dc.isSpike = true;
    };
    lastTotal = dc.total;
    dc.cyear =  y;
    dc.cmonth = m;   
  }  
  catch (JSONException e) {  
    println (e.toString());  
  } 
  
  return(dc);
};

/*
loads data from xml files from usaspending.gov
formats data into SpendingData objects
it may look rocky, but it's just tedious to load xml data
*/
void loadData() {
  data = new ArrayList();
  allProducts = new ArrayList();
  allContractors = new ArrayList();
  for(int i = 0; i < agencies.length; i++) {
    SpendingData sp = new SpendingData();
    XMLElement xml = new XMLElement(this, xmlPath + agencies[i] + ".xml");
    XMLElement dept = xml.getChild("search_criteria").getChild(2); //2nd stores dept
    sp.dept = dept.getStringAttribute("value");
    println("getting data for "+sp.dept);
    XMLElement xData = xml.getChild("data/record");
    XMLElement totals = xData.getChild("totals");
    int fy = int(totals.getChild("fiscal_year").getContent());
    sp.fiscalYear = fy;
    float totalObligation = float(totals.getChild("total_ObligatedAmount").getContent());
    sp.totalObligation = totalObligation;
    float totalPercent = 0;
    try {
      String percent = totals.getChild("percent_of_fiscal_year").getContent();
      totalPercent = float(percent.substring(0,percent.length()-2));
      sp.totalPercent = totalPercent;
    } 
    catch (NullPointerException e) {
      sp.totalPercent = 0; 
    }
    sp.numContractors = int(totals.getChild("number_of_contractors").getContent());
    sp.numTransactions = int(totals.getChild("number_of_transactions").getContent());


    String nameAmtDept = sp.dept;
    XMLElement topProducts = xData.getChild("top_products_or_services_sold");
    for(int j = 0; j < sp.topProducts.length; j++) {
      XMLElement product = topProducts.getChild(j);
      NameAmt prod = new NameAmt();
      prod.dept = sp.dept;
      prod.name = trim(product.getContent());
      prod.amnt = product.getFloatAttribute("total_obligatedAmount");
      sp.topProducts[j] = prod;
      allProducts.add(prod);
    }
    XMLElement topAgencies = xData.getChild("top_contracting_agencies");
    for(int j = 0; j < topAgencies.getChildCount(); j++) {
      XMLElement product = topAgencies.getChild(j);
      NameAmt prod = new NameAmt();
      prod.dept = sp.dept;
      prod.name = trim(product.getContent());
      prod.amnt = product.getFloatAttribute("total_obligatedAmount");
      sp.topAgencies[j] = prod;
    }
    XMLElement topContractors = xData.getChild("top_contractor_parent_companies");
    for(int j = 0; j < sp.topContractors.length; j++) {
      XMLElement product = topContractors.getChild(j);
      NameAmt prod = new NameAmt();
      prod.dept = sp.dept;
      prod.name = trim(product.getContent());
      prod.amnt = product.getFloatAttribute("total_obligatedAmount");
      sp.topContractors[j] = prod;
      allContractors.add(prod);
    }
    data.add(sp);
  }
  Comparator c = new SpendingDataComparator();
  Collections.sort(data,c);
  Comparator c1 = new NameAmtComparator();
  Collections.sort(allProducts,c1); //sort the data in descending order
  Collections.sort(allContractors,c1);
}
